import json
import unittest

from nflxprofile import nflxprofile_pb2
from nflxprofile.flamegraph import NodeJsPackageStackProcessor, get_flame_graph


class TestNodeJsPackageStackProcessor(unittest.TestCase):

    def test_nodejs1(self):
        profile = nflxprofile_pb2.Profile()
        with open("test/fixtures/nodejs1.nflxprofile", "rb") as f:
            profile.ParseFromString(f.read())

        expected = None
        with open("test/fixtures/nodejs1-package.json", "r") as f:
            expected = json.loads(f.read())

        fg = get_flame_graph(profile, None,
                             stack_processor=NodeJsPackageStackProcessor)
        self.assertDictEqual(fg, expected)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: nflxprofile.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='nflxprofile.proto',
  package='nflxprofile',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n\x11nflxprofile.proto\x12\x0bnflxprofile\"7\n\x04\x46ile\x12\x11\n\tfile_name\x18\x01 \x02(\t\x12\x0c\n\x04line\x18\x02 \x01(\r\x12\x0e\n\x06\x63olumn\x18\x03 \x01(\r\"U\n\nStackFrame\x12\x15\n\rfunction_name\x18\x01 \x02(\t\x12\x0f\n\x07libtype\x18\x02 \x01(\t\x12\x1f\n\x04\x66ile\x18\x03 \x01(\x0b\x32\x11.nflxprofile.File\"\xc1\x05\n\x07Profile\x12\x12\n\nstart_time\x18\x01 \x02(\x01\x12\x10\n\x08\x65nd_time\x18\x02 \x02(\x01\x12\x13\n\x07samples\x18\x03 \x03(\rB\x02\x10\x01\x12\x17\n\x0btime_deltas\x18\x04 \x03(\x01\x42\x02\x10\x01\x12.\n\x05nodes\x18\x05 \x03(\x0b\x32\x1f.nflxprofile.Profile.NodesEntry\x12\r\n\x05title\x18\x06 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x07 \x01(\t\x12\x30\n\x06params\x18\x08 \x03(\x0b\x32 .nflxprofile.Profile.ParamsEntry\x12\x17\n\x0bsamples_cpu\x18\t \x03(\rB\x02\x10\x01\x12\x17\n\x0bsamples_pid\x18\n \x03(\rB\x02\x10\x01\x12\x17\n\x0bsamples_tid\x18\x0b \x03(\rB\x02\x10\x01\x12\x19\n\rsamples_value\x18\x0c \x03(\x04\x42\x02\x10\x01\x12\x19\n\x11idle_sample_count\x18\r \x01(\r\x1a\xe2\x01\n\x04Node\x12\x15\n\rfunction_name\x18\x01 \x02(\t\x12\x11\n\thit_count\x18\x02 \x02(\r\x12\x10\n\x08\x63hildren\x18\x03 \x03(\r\x12\x0f\n\x07libtype\x18\x04 \x01(\t\x12\x0e\n\x06parent\x18\x05 \x01(\r\x12\x0b\n\x03pid\x18\x06 \x01(\r\x12\x0b\n\x03tid\x18\x07 \x01(\r\x12\x0b\n\x03\x63pu\x18\x08 \x01(\r\x12\r\n\x05value\x18\t \x01(\x04\x12&\n\x05stack\x18\n \x03(\x0b\x32\x17.nflxprofile.StackFrame\x12\x1f\n\x04\x66ile\x18\x0b \x01(\x0b\x32\x11.nflxprofile.File\x1aG\n\nNodesEntry\x12\x0b\n\x03key\x18\x01 \x01(\r\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.nflxprofile.Profile.Node:\x02\x38\x01\x1a-\n\x0bParamsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01')
)




_FILE = _descriptor.Descriptor(
  name='File',
  full_name='nflxprofile.File',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='file_name', full_name='nflxprofile.File.file_name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='line', full_name='nflxprofile.File.line', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='column', full_name='nflxprofile.File.column', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=34,
  serialized_end=89,
)


_STACKFRAME = _descriptor.Descriptor(
  name='StackFrame',
  full_name='nflxprofile.StackFrame',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='function_name', full_name='nflxprofile.StackFrame.function_name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='libtype', full_name='nflxprofile.StackFrame.libtype', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file', full_name='nflxprofile.StackFrame.file', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=91,
  serialized_end=176,
)


_PROFILE_NODE = _descriptor.Descriptor(
  name='Node',
  full_name='nflxprofile.Profile.Node',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='function_name', full_name='nflxprofile.Profile.Node.function_name', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hit_count', full_name='nflxprofile.Profile.Node.hit_count', index=1,
      number=2, type=13, cpp_type=3, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='children', full_name='nflxprofile.Profile.Node.children', index=2,
      number=3, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='libtype', full_name='nflxprofile.Profile.Node.libtype', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='parent', full_name='nflxprofile.Profile.Node.parent', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pid', full_name='nflxprofile.Profile.Node.pid', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tid', full_name='nflxprofile.Profile.Node.tid', index=6,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cpu', full_name='nflxprofile.Profile.Node.cpu', index=7,
      number=8, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='nflxprofile.Profile.Node.value', index=8,
      number=9, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='stack', full_name='nflxprofile.Profile.Node.stack', index=9,
      number=10, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='file', full_name='nflxprofile.Profile.Node.file', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=538,
  serialized_end=764,
)

_PROFILE_NODESENTRY = _descriptor.Descriptor(
  name='NodesEntry',
  full_name='nflxprofile.Profile.NodesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='nflxprofile.Profile.NodesEntry.key', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='nflxprofile.Profile.NodesEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=766,
  serialized_end=837,
)

_PROFILE_PARAMSENTRY = _descriptor.Descriptor(
  name='ParamsEntry',
  full_name='nflxprofile.Profile.ParamsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='nflxprofile.Profile.ParamsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='nflxprofile.Profile.ParamsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=839,
  serialized_end=884,
)

_PROFILE = _descriptor.Descriptor(
  name='Profile',
  full_name='nflxprofile.Profile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='start_time', full_name='nflxprofile.Profile.start_time', index=0,
      number=1, type=1, cpp_type=5, label=2,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='nflxprofile.Profile.end_time', index=1,
      number=2, type=1, cpp_type=5, label=2,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='samples', full_name='nflxprofile.Profile.samples', index=2,
      number=3, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\020\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='time_deltas', full_name='nflxprofile.Profile.time_deltas', index=3,
      number=4, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\020\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='nodes', full_name='nflxprofile.Profile.nodes', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='title', full_name='nflxprofile.Profile.title', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='nflxprofile.Profile.description', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='params', full_name='nflxprofile.Profile.params', index=7,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='samples_cpu', full_name='nflxprofile.Profile.samples_cpu', index=8,
      number=9, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\020\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='samples_pid', full_name='nflxprofile.Profile.samples_pid', index=9,
      number=10, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\020\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='samples_tid', full_name='nflxprofile.Profile.samples_tid', index=10,
      number=11, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\020\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='samples_value', full_name='nflxprofile.Profile.samples_value', index=11,
      number=12, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\020\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='idle_sample_count', full_name='nflxprofile.Profile.idle_sample_count', index=12,
      number=13, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_PROFILE_NODE, _PROFILE_NODESENTRY, _PROFILE_PARAMSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=179,
  serialized_end=884,
)

_STACKFRAME.fields_by_name['file'].message_type = _FILE
_PROFILE_NODE.fields_by_name['stack'].message_type = _STACKFRAME
_PROFILE_NODE.fields_by_name['file'].message_type = _FILE
_PROFILE_NODE.containing_type = _PROFILE
_PROFILE_NODESENTRY.fields_by_name['value'].message_type = _PROFILE_NODE
_PROFILE_NODESENTRY.containing_type = _PROFILE
_PROFILE_PARAMSENTRY.containing_type = _PROFILE
_PROFILE.fields_by_name['nodes'].message_type = _PROFILE_NODESENTRY
_PROFILE.fields_by_name['params'].message_type = _PROFILE_PARAMSENTRY
DESCRIPTOR.message_types_by_name['File'] = _FILE
DESCRIPTOR.message_types_by_name['StackFrame'] = _STACKFRAME
DESCRIPTOR.message_types_by_name['Profile'] = _PROFILE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

File = _reflection.GeneratedProtocolMessageType('File', (_message.Message,), dict(
  DESCRIPTOR = _FILE,
  __module__ = 'nflxprofile_pb2'
  # @@protoc_insertion_point(class_scope:nflxprofile.File)
  ))
_sym_db.RegisterMessage(File)

StackFrame = _reflection.GeneratedProtocolMessageType('StackFrame', (_message.Message,), dict(
  DESCRIPTOR = _STACKFRAME,
  __module__ = 'nflxprofile_pb2'
  # @@protoc_insertion_point(class_scope:nflxprofile.StackFrame)
  ))
_sym_db.RegisterMessage(StackFrame)

Profile = _reflection.GeneratedProtocolMessageType('Profile', (_message.Message,), dict(

  Node = _reflection.GeneratedProtocolMessageType('Node', (_message.Message,), dict(
    DESCRIPTOR = _PROFILE_NODE,
    __module__ = 'nflxprofile_pb2'
    # @@protoc_insertion_point(class_scope:nflxprofile.Profile.Node)
    ))
  ,

  NodesEntry = _reflection.GeneratedProtocolMessageType('NodesEntry', (_message.Message,), dict(
    DESCRIPTOR = _PROFILE_NODESENTRY,
    __module__ = 'nflxprofile_pb2'
    # @@protoc_insertion_point(class_scope:nflxprofile.Profile.NodesEntry)
    ))
  ,

  ParamsEntry = _reflection.GeneratedProtocolMessageType('ParamsEntry', (_message.Message,), dict(
    DESCRIPTOR = _PROFILE_PARAMSENTRY,
    __module__ = 'nflxprofile_pb2'
    # @@protoc_insertion_point(class_scope:nflxprofile.Profile.ParamsEntry)
    ))
  ,
  DESCRIPTOR = _PROFILE,
  __module__ = 'nflxprofile_pb2'
  # @@protoc_insertion_point(class_scope:nflxprofile.Profile)
  ))
_sym_db.RegisterMessage(Profile)
_sym_db.RegisterMessage(Profile.Node)
_sym_db.RegisterMessage(Profile.NodesEntry)
_sym_db.RegisterMessage(Profile.ParamsEntry)


_PROFILE_NODESENTRY._options = None
_PROFILE_PARAMSENTRY._options = None
_PROFILE.fields_by_name['samples']._options = None
_PROFILE.fields_by_name['time_deltas']._options = None
_PROFILE.fields_by_name['samples_cpu']._options = None
_PROFILE.fields_by_name['samples_pid']._options = None
_PROFILE.fields_by_name['samples_tid']._options = None
_PROFILE.fields_by_name['samples_value']._options = None
# @@protoc_insertion_point(module_scope)



"""Flame graph module for generating flame graphs from nflxprofile profiles."""

__ALL__ = ['get_flame_graph', 'StackProcessor', 'NodeJsStackProcessor', 'NodeJsPackageStackProcessor']

import math
import os
import pathlib

from nflxprofile import nflxprofile_pb2


def _get_child(node, frame):
    """Docstring for public method."""
    if isinstance(frame, dict):
        name = frame.get('name', "")
        libtype = frame.get("libtype", "")
        filename = frame.get('extras', {}).get('file', "")
    else:
        name = frame.function_name
        libtype = frame.libtype
        filename = frame.file.file_name or ""
        if filename:
            filename = "%s:%d" % (filename, frame.file.line or 0)
    for child in node['children']:
        if child['name'] == name and child['libtype'] == libtype:
            child_file = child.get('extras', {}).get('file', "")
            if filename == child_file:
                return child
    return None


def _generate_regular_stacks(nflxprofile_nodes, root_node_id):
    stacks = {}
    queue = []
    queue.append((root_node_id, None))

    while queue:
        (nflxprofile_node_id, parent_node_id) = queue.pop(0)
        nflxprofile_node = nflxprofile_nodes[nflxprofile_node_id]
        stack_frame = nflxprofile_pb2.StackFrame()
        stack_frame.function_name = nflxprofile_node.function_name
        stack_frame.libtype = nflxprofile_node.libtype
        if not parent_node_id:
            stacks[nflxprofile_node_id] = [stack_frame]
        else:
            stacks[nflxprofile_node_id] = stacks[parent_node_id] + [stack_frame]
        for child_id in nflxprofile_node.children:
            queue.append((child_id, nflxprofile_node_id))

    return stacks


def _generate_package_name_stacks(nflxprofile_nodes):
    stacks = {}
    for key in nflxprofile_nodes:
        nflxprofile_node = nflxprofile_nodes[key]
        function_name = nflxprofile_node.function_name.split(';')[0]
        function_name_arr = function_name.split('/')
        stack_arr = []
        for package_name in function_name_arr:
            stack_frame = nflxprofile_pb2.StackFrame()
            stack_frame.function_name = package_name
            stack_frame.libtype = nflxprofile_node.libtype
            stack_arr.append(stack_frame)
        stacks[key] = stack_arr

    return stacks


def _generate_stacks(nflxprofile_nodes, root_node_id, package_name=False):
    if package_name:
        return _generate_package_name_stacks(nflxprofile_nodes)
    return _generate_regular_stacks(nflxprofile_nodes, root_node_id)


def _get_stack(nflxprofile_nodes, node_id, has_node_stack=False, pid_comm=None, **args):
    """Get node stack using parent pointers or predefined stack."""
    inverted = args.get("inverted", False)
    package_name = args.get("package_name", False)

    stack = []

    # package name, only need first node
    if package_name:
        function_name = nflxprofile_nodes[node_id].function_name
        if has_node_stack:
            # uses node stack format, can't use node's function name
            node_stack = nflxprofile_nodes[node_id].stack
            function_name = node_stack[-1].function_name
        sanitized_function_name = function_name.split(';')[0]
        function_name_arr = sanitized_function_name.split('/')
        for name in function_name_arr:
            stack_frame = nflxprofile_pb2.StackFrame()
            stack_frame.function_name = name
            stack_frame.libtype = nflxprofile_nodes[node_id].libtype
            stack.append(stack_frame)
        if inverted:
            return reversed(stack)
        return stack

    # has node stack calculated, returning that
    if has_node_stack:
        function_name = nflxprofile_nodes[node_id].function_name
        pid = nflxprofile_nodes[node_id].pid
        if pid_comm and pid and pid in pid_comm:
            function_name = pid_comm[pid]
        stack_frame = nflxprofile_pb2.StackFrame()
        stack_frame.function_name = function_name
        stack_frame.libtype = nflxprofile_nodes[node_id].libtype
        stack = [stack_frame] + list(nflxprofile_nodes[node_id].stack)
        if inverted:
            return reversed(stack)
        return stack

    # need to use parent id
    nflxprofile_node_id = node_id
    while True:
        nflxprofile_node = nflxprofile_nodes[nflxprofile_node_id]
        if inverted:
            stack.append((nflxprofile_node.function_name, nflxprofile_node.libtype))
        else:
            stack_frame = nflxprofile_pb2.StackFrame()
            stack_frame.function_name = nflxprofile_node.function_name
            stack_frame.libtype = nflxprofile_node.libtype
            stack.insert(
                0, stack_frame)
        if not nflxprofile_nodes[nflxprofile_node_id].parent:
            break
        nflxprofile_node_id = nflxprofile_node.parent
    return stack


class FrameExtras:
    """Generic class to store extra information about a stack frame."""

    def __init__(self):
        """Constructor."""
        self.v8_jit = False
        self.javascript = False
        self.real_name = ""
        self.optimized = None

    def __repr__(self):
        return ("FrameExtras(v8_jit=%s, javascript=%s, real_name=%s, optimized=%s)"
                % (self.v8_jit, self.javascript, self.real_name, self.optimized))


class Frame:
    def __init__(self, frame):
        self.frame = frame

    def __getattr__(self, name):
        return getattr(self.frame, name)


class StackProcessor:
    """Processes a stack trace, extend it to add custom processing."""

    def __init__(self, root, profile, index, value=1):
        """Constructor."""
        self.current_node = root
        self.value = value
        self.empty_extras = FrameExtras()

    def process_frame(self, frame):
        """Process one frame, returning the processed frame plus extras."""
        return frame, self.empty_extras

    # pylint: disable=no-self-use
    def should_skip_frame(self, frame, frame_extras):
        """Check if this frame should be skipped."""
        return False

    def process_extras(self, child, frame, frame_extras):
        """Process extras and save it in the given node."""
        extras = child.get('extras', {})

        if frame.file.file_name:
            extras = child.get('extras', {})
            if 'file' not in extras:
                extras['file'] = frame.file.file_name
                if extras['file']:
                    extras['file'] = ('%s:%d' % (extras['file'], frame.file.line))
            child['extras'] = extras
        child['extras'] = extras

    def process(self, stack):
        """Processes a stack trace.

        You probably want to avoid overriding this method. Override other
        methods to customize behavior instead.
        """
        for frame in stack:
            frame, frame_extras = self.process_frame(frame)
            if self.should_skip_frame(frame, frame_extras):
                continue
            child = _get_child(self.current_node, frame)
            if child is None:
                child = {
                    'name': frame.function_name,
                    'libtype': frame.libtype,
                    'value': 0,
                    'children': []
                }
                self.current_node['children'].append(child)
            self.process_extras(child, frame, frame_extras)
            self.current_node = child
        self.current_node['value'] = self.current_node['value'] + self.value


class NodeJsPackageStackProcessor(StackProcessor):

    def __init__(self, root, profile, index, value=1):
        """Constructor."""
        super().__init__(root, profile, index, value)
        self.current_package = None
        self.packages_cache = {}

    def get_package(self, frame):
        name = frame.function_name
        if name in self.packages_cache:
            return self.packages_cache[name]

        package = None
        if name.startswith("LazyCompile:") or name.startswith("InterpretedFunction:"):
            name = name[name.index(":") + 1:]
            if name and name[0] == '*':
                name = name[1:]

            if " " in name:
                package = name[name.index(" ") + 1:]

        if package is not None:
            if ":" in package:
                package = package.rsplit(":", 1)[0]
            if "node_modules" in package:
                package = pathlib.Path(package.rsplit("node_modules", 1)[1])
                if package.parts[1].startswith("@"):
                    package = os.path.join(*package.parts[1:3])
                else:
                    package = package.parts[1]
            elif package.startswith("/") or "[eval" in package:
                return "(app code)"
            else:
                package = "(node api)"
        else:
            if frame.libtype == 'kernel':
                return '(kernel)'
            else:
                return '(native)'

        return package

    def should_skip(self, name):
        # We'll skip for known, non-expensive builtins which can appear between
        # JS frames. Showing those would fragment the FlameGraph unecessarily.
        if 'ArgumentsAdaptorTrampoline' in name:
            return True
        if name.startswith('Builtin'):
            if 'Construct' in name:
                return True
            if 'LoadIC' in name or 'StoreIC' in name:
                return True
            if 'InterpreterEntryTrampoline' in name:
                return True
        if name.startswith('BytecodeHandler'):
            return True
        return False

    def process(self, stack):
        # We always start with native
        current_frame = nflxprofile_pb2.StackFrame()
        current_frame.function_name = "(native)"
        current_frame.libtype = ""
        processed_stack = []
        current_stack = []
        for frame in stack:
            package = self.get_package(frame)

            if package == current_frame.function_name or self.should_skip(frame.function_name):
                current_stack.append(frame)
                continue

            processed_stack.append(current_frame)

            current_frame = nflxprofile_pb2.StackFrame()
            current_frame.function_name = package
            current_frame.libtype = ""
            current_stack = []

        processed_stack.append(current_frame)

        return super().process(processed_stack)


class NodeJsStackProcessor(StackProcessor):
    """Node.js mode stack processor.

    Sanitize JIT function names, extract file name from frame name, group
    interpreted and compiled function, flag V8 builtins, hide
    ArgumentsAdaptorTrampoline frames (but store count in the next frame so we
    can exhibit this information on the interface).
    """

    def __init__(self, root, profile, index, value=1):
        """Constructor."""
        super().__init__(root, profile, index, value)
        self.argument_adaptor = None

    def should_skip_frame(self, frame, frame_extras):
        """Skip ArgumentsAdaptorTrampoline.

        ArgumentsAdaptorTrampoline frames are inserted between calls when the
        caller calls the callee with wrong signature. While this information is
        relevant (since ArgumentsAdaptorTrampoline is not free), it can mess
        with frame grouping. Skipping this frame makes sense and we can show
        information about it in the following frame when hovering it.
        """
        if "ArgumentsAdaptorTrampoline" in frame.function_name:
            self.argument_adaptor = self.value
            return True
        return False

    def process_extras(self, child, frame, frame_extras):
        """Add Node.js specific extras.

        Add % of times a function was called with mismatched arguments, % of
        times it executed JIT instead of intepreted, as well as some metadata
        used for coloring the flamegraph.
        """
        extras = child.get('extras', {'optimized': 0})
        extras['javascript'] = frame_extras.javascript
        extras['v8_jit'] = frame_extras.v8_jit
        extras['optimized'] = extras['optimized'] + (frame_extras.optimized and self.value or 0)
        extras['realName'] = frame_extras.real_name
        if self.argument_adaptor:
            extras['argumentAdaptor'] = extras.get('argumentAdaptor', 0) + self.argument_adaptor
            self.argument_adaptor = None
        child['extras'] = extras
        super().process_extras(child, frame, frame_extras)

    def process_frame(self, frame):
        """Process frame.

        Sanitize JIT function names, extract file name from frame name,
        generate some metadata used by other methods.
        """
        processed_frame = Frame(frame)
        frame_extras = FrameExtras()
        frame_extras.v8_jit = False
        frame_extras.javascript = False
        name = frame.function_name

        frame_extras.real_name = name

        if name.startswith("LazyCompile:") or name.startswith("InterpretedFunction:"):
            frame_extras.v8_jit = True
            frame_extras.javascript = True
            frame_extras.optimized = name.startswith("LazyCompile:")

            name = name[name.index(":") + 1:]
            if name and name[0] == '*':
                name = name[1:]

            if " " in name:
                frame.file.file_name = name[name.index(" ") + 1:]
                if frame.file.file_name:
                    name = name[:name.index(" ")]
            processed_frame.function_name = name or "(anonymous)"

        return processed_frame, frame_extras


class SampleFilter:
    """Interface for sample filters.

    Extend this class and override should_skip to create a new filter.
    """

    # pylint: disable=unused-argument
    def __init__(self, profile, **args):
        """Default constructor for a filter."""
        self.profile = profile

    # pylint: disable=no-self-use
    def should_skip(self, sample, index, current_time):
        """Returns false if a given sample shouldn't be processed."""
        return False


class RangeSampleFilter(SampleFilter):
    """Filter all samples within a given range."""

    def __init__(self, profile, range_start=None, range_end=None, **args):
        """Range filter constructor."""
        super().__init__(profile)
        if range_start is None or range_end is None:
            self.range_start = self.range_end = None
            return
        start_time = math.floor(profile.start_time)
        self.range_start = (start_time + range_start)
        self.range_end = (start_time + range_end)

    def should_skip(self, sample, index, current_time):
        """Returns false if a given sample is not in range."""
        if self.range_start is None or self.range_end is None:
            return False
        return not self.range_start <= current_time < self.range_end


class CPUSampleFilter(SampleFilter):
    """Filter all samples for a given CPU."""

    def __init__(self, profile, cpu=None, **args):
        """CPU filter constructor."""
        super().__init__(profile)
        self.cpu = cpu
        self.samples_cpu = None

        if 'has_samples_cpu' not in profile.params:
            return
        if profile.params['has_samples_cpu'] != 'true':
            return

        self.samples_cpu = profile.samples_cpu

    def should_skip(self, sample, index, current_time):
        """Returns false if a given sample was not running on this CPU."""
        if self.cpu is None or self.samples_cpu is None:
            return False
        return self.cpu != self.samples_cpu[index]


class PIDSampleFilter(SampleFilter):
    """Filter all samples for a given PID."""

    def __init__(self, profile, pid=None, **args):
        """PID filter constructor."""
        super().__init__(profile)
        self.pid = pid
        self.samples_pid = None

        if 'has_samples_pid' not in profile.params:
            return
        if profile.params['has_samples_pid'] != 'true':
            return

        self.samples_pid = profile.samples_pid

    def should_skip(self, sample, index, current_time):
        """Returns false if a given sample doesn't include PID."""
        if self.pid is None or self.samples_pid is None:
            return False
        return self.pid != self.samples_pid[index]


class TIDSampleFilter(SampleFilter):
    """Filter all samples for a given TID."""

    def __init__(self, profile, tid=None, **args):
        """PID filter constructor."""
        super().__init__(profile)
        self.tid = tid
        self.samples_tid = None

        if 'has_samples_tid' not in profile.params:
            return
        if profile.params['has_samples_tid'] != 'true':
            return

        self.samples_tid = profile.samples_tid

    def should_skip(self, sample, index, current_time):
        """Returns false if a given sample doesn't include TID."""
        if self.tid is None or self.samples_tid is None:
            return False
        return self.tid != self.samples_tid[index]


# pylint: disable=too-many-locals
# pylint: disable=too-many-branches
# pylint: disable=too-many-statements
def get_flame_graph(profile, pid_comm, **args):
    """Generate flame graph from a nflxprofile profile."""
    inverted = args.get("inverted", False)
    package_name = args.get("package_name", False)
    use_sample_value = args.get("use_sample_value", False)
    stack_processor_class = args.get("stack_processor", StackProcessor)

    nodes = profile.nodes
    root_id = 0

    samples = profile.samples
    time_deltas = profile.time_deltas
    start_time = profile.start_time

    sample_filters = [
        RangeSampleFilter(profile, **args),
        CPUSampleFilter(profile, **args),
        PIDSampleFilter(profile, **args),
        TIDSampleFilter(profile, **args)
    ]

    root = {
        'name': 'root',
        'libtype': '',
        'value': 0,
        'children': []
    }
    current_time = start_time + time_deltas[0]

    has_node_stack = \
        'has_node_stack' in profile.params and profile.params['has_node_stack'] == 'true'
    has_parent = \
        'has_parent' in profile.params and profile.params['has_parent'] == 'true'

    samples_value = None
    if 'hasValues' in profile.params and profile.params['hasValues'] == 'true':
        samples_value = profile.samples_value

    stacks = None
    if (not has_node_stack) and (not has_parent):
        # don't have stacks or parent pointer, generating stacks manually
        # case for very old nflxprofile
        stacks = _generate_stacks(nodes, root_id, package_name)

    for index, sample in enumerate(samples):
        if index == (len(samples) - 1):  # last sample
            break
        delta = time_deltas[index + 1]
        current_time += delta

        should_skip = False
        for sample_filter in sample_filters:
            should_skip = sample_filter.should_skip(sample, index, current_time)
            if should_skip:
                break
        if should_skip:
            continue

        sample_value = 1
        if use_sample_value:
            sample_value = samples_value[index] if samples_value else None
        stack_processor = stack_processor_class(root, profile, index, sample_value)

        if stacks:
            stack = stacks[sample] if not inverted else reversed(stacks[sample])
        else:
            stack = _get_stack(nodes, sample, has_node_stack, pid_comm, **args)

        stack_processor.process(stack)
    return root



import argparse
import functools
import json

from nflxprofile.convert.v8_cpuprofile import parse as v8_parse
from nflxprofile.flamegraph import NodeJsPackageStackProcessor, NodeJsStackProcessor, StackProcessor, get_flame_graph
from nflxprofile.nflxprofile_pb2 import Profile

STACK_PROCESSOR = {
    'default':  StackProcessor,
    'nodejs':  NodeJsStackProcessor,
    'nodejs-package':  NodeJsPackageStackProcessor,
}


def get_input_format(input_format, input_files=[]):
    if input_format is None:
        if functools.reduce(lambda a, b: a and b.endswith(".cpuprofile"), input_files, True):
            input_format = 'v8'
        elif len(input_files) != 1:
            raise ValueError("Unable to infer input type. Please use --input-format")
        elif input_files[0].endswith(".nflxprofile"):
            input_format = 'nflxprofile'
        else:
            input_format = 'perf'
    return input_format


def get_output_format(output_format, output_file=None):
    if output_format is None:
        if output_file is None or output_file.endswith(".nflxprofile"):
            output_format = 'nflxprofile'
        elif output_file.endswith(".json"):
            output_format = 'tree'
        else:
            raise ValueError("Unable to infer output type. Please use --output-format")
    return output_format


def validate_input_output(input_format, output_format, input_files=[]):
    if input_format == 'nflxprofile':
        if output_format != 'tree':
            raise ValueError("Can't convert %s to %s" % (input_format, output_format))
    else:
        if output_format != 'nflxprofile':
            raise ValueError("Can't convert %s to %s" % (input_format, output_format))

    if input_format != 'v8' and len(input_files) > 1:
        raise ValueError("Only V8 .cpuprofile support multiple input files")


def main():
    parser = argparse.ArgumentParser(prog="nflxprofile", description=('Parse '
                                     'common profile/tracing formats into nflxprofile'))
    parser.add_argument('--output')
    parser.add_argument('--input-format', choices=['v8', 'perf', 'nflxprofile'])
    parser.add_argument('--output-format', choices=['nflxprofile', 'tree'])
    parser.add_argument('--force', action="store_true")
    parser.add_argument('--extra-options', type=json.loads)
    parser.add_argument('input', nargs="+")

    args = parser.parse_args()

    input_format = get_input_format(args.input_format, args.input)
    output_format = get_output_format(args.output_format, args.output)

    validate_input_output(input_format, output_format, args.input)

    if output_format == 'nflxprofile':
        out = args.output
        if not out:
            out = 'profile.nflxprofile'

        filenames = args.input
        extra_options = args.extra_options or {}

        profile = None
        if input_format == 'v8':
            profiles = []
            for filename in filenames:
                with open(filename, 'r') as f:
                    profiles.append(json.loads(f.read()))
            profile = v8_parse(profiles, **extra_options)

        with open(out, 'wb') as f:
            f.write(profile.SerializeToString())

    elif output_format == 'tree':
        out = args.output
        if not out:
            out = 'profile.json'

        filename = args.input[0]
        extra_options = args.extra_options or {}

        extra_options['stack_processor'] = STACK_PROCESSOR[extra_options.get('stack_processor', 'default')]

        tree = {}
        if input_format == 'nflxprofile':
            profile = Profile()
            with open(filename, 'rb') as f:
                profile.ParseFromString(f.read())
            tree = get_flame_graph(profile, {}, **extra_options)

        with open(out, 'w') as f:
            f.write(json.dumps(tree))
